/*

MANGLER - by Lou Duchez

Mangles the word order of documents.

SYNTAX: mangler [infile] [outfile] [-s<number>]

All parameters are optional; if "infile" or "outfile" is not specified,
the program uses standard input or output.

The "-s" option allows you to select a "seed" for the random number
generator, from 1 to 2147483647, altering the sequence of random numbers
generated.  Example: "-s3948" sets the seed to 3948.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>

#define STRINSIZE 0x3000
#define WRAPPOS 72

#define BOOLEAN char
#define TRUE 1
#define FALSE 0

#define DEFSEED1 593832
#define DEFSEED2 3282394

#define ALPHAMAX 256

struct wprec              /* Stores Instances of words from text */
{
  struct wprec *next, *nextalpha; /* next word in doc; next instance of word */
  struct tabrec *trec;            /* points to "Main" record */
};

struct tabrec             /* "Main" word record -- tabulates word behavior */
{
  struct wprec *first, *last;     /* first Instance, last Instance */
  struct tabrec *next;            /* next Main record */
  char *word;                     /* stores the text of the word */
  unsigned int wordcnt;           /* How many Instances of word exist */
};

char delims[] = " \t\n\015\012\011";    /* word delimiters to honor */
char outbuffer[200];                    /* for spewing output */
struct wprec *firstwprec, *lastwprec;   /* beginning & end Instances */
struct tabrec *firstalpha[ALPHAMAX],    /* stores beginning and end Main */
              *lastalpha[ALPHAMAX];     /*  records that start with a letter */

long int seed1, seed2;            /* used in random number generator */
/*
--------------------------------------------
*/
unsigned int lrand(unsigned int maxrand)

/* Generate random number from 0 to maxrand - 1 */

{
  long int zz, kk;
  
  if (maxrand <= 1) return 0;  /* Only one possible return value --
                                  return it without altering seeds */

   kk = seed1 / 53668;
   seed1 = 40014*(seed1 - kk*53668) - kk*12211;
   if (seed1 < 0) seed1 += 2147483563;

   kk = seed2 / 52774;
   seed2 = 40692*(seed2 - kk*52774) - kk*3791;
   if (seed2 < 0) seed2 += 2147483399;

   zz = seed1 + seed2;

   if (zz < 1) zz += 2147483562;

   return zz % maxrand;
}
/*
--------------------------------------------
*/
void memerr()

/* Display memory error if there's not enough heap space */

{
  printf("Error: cannot allocate enough memory to run program.\n");
  exit(1);
}
/*
--------------------------------------------
*/
void storeword(char *newword)

/* Stores instance of word to "Instances" linked list, and creates a
   "Main" linked list entry if required */

{
  struct wprec *thiswprec;
  struct tabrec *thistabrec;
  char thischr;
  
  if (newword[0] == '\0') return;             /* Don't store nulls */
  
  thischr = newword[0];                       /* get starting letter */
  
  /* Look for an existing Main record for this word */
  
  thistabrec = firstalpha[thischr];
  while ((thistabrec != NULL) && (strcmp((*thistabrec).word, newword) != 0))
        thistabrec = (*thistabrec).next;
  
  /* no Main record found -- add one */
  
  if (thistabrec == NULL)
  {
    
    /* allocate memory for Main record */
  
    thistabrec = malloc(sizeof(struct tabrec));   /* Allocate mem for struct */
    if (thistabrec == NULL) memerr();
    
    /* put Main record in linked list for that starting letter */
    
    if (firstalpha[thischr] == NULL) firstalpha[thischr] = thistabrec;
    if (lastalpha[thischr] != NULL) (*lastalpha[thischr]).next = thistabrec;
    lastalpha[thischr] = thistabrec;
    
    /* initialize pointers to Instances, as NULL */

    (*thistabrec).first = NULL;
    (*thistabrec).last = NULL;

    /* set pointer to next Main entry, as NULL */

    (*thistabrec).next = NULL;

    /* zero Instances of this word have been identified so far */

    (*thistabrec).wordcnt = 0;
    
    /* store string */
    
    (*thistabrec).word = malloc(strlen(newword) + 1);
    if ((*thistabrec).word == NULL) memerr();
    strcpy((*thistabrec).word, newword);
    
    /* new Main record has been created */
    
  }
  
  /* increment Instance counter */
  
  (*thistabrec).wordcnt++;
  
  /* allocate memory for this Instance */
  
  thiswprec = malloc(sizeof(struct wprec));
  if (thiswprec == NULL) memerr();
  
  /* initialize pointers to next word in doc, and next Instance of this
     word, to NULL */
  
  (*thiswprec).next = NULL;
  (*thiswprec).nextalpha = NULL;

  /* put this Instance in the Instance linked list */

  if (firstwprec == NULL) firstwprec = thiswprec;
  if (lastwprec != NULL) (*lastwprec).next = thiswprec;
  
  /* mark this Instance as the last Instance stored */
  
  lastwprec = thiswprec;
  
  /* set up pointer to the Main record */

  (*thiswprec).trec = thistabrec;
  
  /* update pointers to Instances in the Main record */
  
  if ((*thistabrec).first == NULL) (*thistabrec).first = thiswprec;
  if ((*thistabrec).last != NULL) (*(*thistabrec).last).nextalpha = thiswprec;
  (*thistabrec).last = thiswprec;
}
/*
--------------------------------------------
*/
void storewords(BOOLEAN use_stdin, char *finname)

/* Runs loop to save off all words */

{
  char strin[STRINSIZE + 1], *tokptr;
  FILE *fin;
  int i;
  
  for (i = 0; i < ALPHAMAX; i++)          /* initialize linked lists */
  {                                       /*  for storing Main records */
    firstalpha[i] = NULL;                 /*  by letter */
    lastalpha[i] = NULL;
  }

  firstwprec = NULL;                      /* initialize linked list */
  lastwprec = NULL;                       /*  for storing instances */
  
  if (use_stdin) fin = stdin;
  else
  {
    fin = fopen(finname, "r");            /* Open input file */
    if (fin == NULL)
    {
      printf("Error: could not open %s for input.\n", finname);
      exit(1);
    }
  }
  
  while (!feof(fin))                    /* Loop to read file */
  {
    if (fgets(strin, STRINSIZE, fin) != NULL)  /* Read line */
    {
      tokptr = strtok(strin, delims);     /* Get first word on line */
      while (tokptr != NULL)              /* Loop to process all words */
      {
        storeword(tokptr);                /* Save off this word on line */
        tokptr = strtok(NULL, delims);    /* Get next word on line */
      }
    }
  }
  if (!use_stdin) fclose(fin);
}
/*
--------------------------------------------
*/
void outputtext(FILE *fout, char *newword)      /* write out text */
{
  char strout[200];
  int i;
  
  strcat(outbuffer, " ");               /* add new word to text */
  strcat(outbuffer, newword);
  
  if (strlen(outbuffer) > WRAPPOS)      /* line too long - must wrap */
  {
    strcpy(strout, outbuffer);          /* Find space to wrap line at */
    i = WRAPPOS;
    while ((i >= 0) && (strout[i] != ' ')) i--;
    if (i < 0)                                    /* No good place to wrap */
    {
      strcpy(outbuffer, strout + WRAPPOS - 1);
      strout[WRAPPOS - 1] = '\0';
    }
    else                                          /* Found place to wrap */
    {
      strcpy(outbuffer, strout + i + 1);
      strout[i] = '\0';
    }
    fprintf(fout, "%s\n", strout);           /* print out text */
  }
  
  if (newword[0] == '\0')             /* End of text -- spit out buffer */
     fprintf(fout, "%s\n", outbuffer);
}
/*
--------------------------------------------
*/
void printwords(BOOLEAN use_stdout, char *foutname)  /* Generate output */
{
  struct tabrec *tabptr;
  struct wprec *wpptr;
  int i, semirand;
  FILE *fout;

  if (use_stdout) fout = stdout;  /* output to normal output for piping */
  else
  {
    fout = fopen(foutname, "w");  /* otherwise, open output file */
    if (fout == NULL)
    {
      printf("Error: could not open %s for output.\n", foutname);
      exit(1);
    }
  }

  outbuffer[0] = '\0';            /* Init output buffer */
  wpptr = firstwprec;             /* Go to top of word LL */
  while (wpptr != NULL)           /* Loop while not at end of word LL */
  {
    tabptr = (*wpptr).trec;
    outputtext(fout, (*tabptr).word);    /* Spit out current word */
    semirand = lrand((*tabptr).wordcnt) + 1;
    wpptr = (*tabptr).first;
    for (i = 1; i < semirand; i++) wpptr = (*wpptr).nextalpha;
    wpptr = (*wpptr).next;
  }
  outputtext(fout, "");                 /* Spit out any remaining buffer */
  if (!use_stdout) fclose(fout);
}
/*
--------------------------------------------
*/
void main(int argc, char *argv[])
{
  char finname[200], foutname[200], tmpparam[200];
  BOOLEAN use_stdin, use_stdout;
  int i;
  
  seed1 = DEFSEED1;               /* initialize randomization */
  seed2 = DEFSEED2;
  
  use_stdin = TRUE;               /* defaults: use STDIN and STDOUT */
  use_stdout = TRUE;
  
  for (i = 2; i <= argc; i++)     /* look for "-s" switch for changing seed */
  {
    strcpy(tmpparam, argv[i - 1]);
    if ((tmpparam[0] == '-') || (tmpparam[0] == '/'))
    {
      if (tmpparam[1] == '?')
      {
        printf("MANGLER - by Lou Duchez\n\n");
        printf("Mangles the word order of documents.\n\n");
        printf("SYNTAX: mangler [infile] [outfile] [-s<number>]\n\n");
        printf("All parameters are optional; if \"infile\" or \"outfile\" is not specified,\n");
        printf("the program uses standard input or output.\n\n");
        printf("The \"-s\" option allows you to select a \"seed\" for the random number\n");
        printf("generator, from 1 to 2147483647, altering the sequence of random numbers\n");
        printf("generated.  Example: \"-s3948\" sets the seed to 3948.\n");
        exit(0);
      }
      if ((tmpparam[1] == 's') || (tmpparam[1] == 'S'))
      {
        sscanf(tmpparam + 2, "%ld", &seed1);
        if (seed1 <= 0) seed1 = DEFSEED1;
      }
    }
    else if (use_stdin)           /* this param must be an input file name */
    {
      use_stdin = FALSE;
      strcpy(finname, tmpparam);
    }
    else if (use_stdout)          /* this param must be an output file name */
    {
      use_stdout = FALSE;
      strcpy(foutname, tmpparam);
    }
  }
  storewords(use_stdin, finname);     /* store words off to linked list */
  printwords(use_stdout, foutname);   /* print words out in mangled order */
}
